/*
 * (C) Copyright IBM Corp. 2003    All Rights Reserved.
 * mdstat.c ($Revision: 1.5 $)
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corp.
 *
 * Author(s): S. Bader <shbader@de.ibm.com>
 *
 * This file is part of MDMPD.
 *
 * MDMPD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MDMPD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MDMPD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <sys/poll.h>
#include <sys/stat.h>

#include <linux/kdev_t.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <mdstat.h>

#define ERR(t, ...)	fprintf(stderr, t, ## __VA_ARGS__)
#if DEBUG > 4
#define DBG(t, ...)	fprintf(stderr, "%d: " t, __LINE__ , ## __VA_ARGS__)
#else
#define DBG(t, ...)
#endif

#define BUFLEN	160
#define DEVDIR	"/dev"

static char *	revision	= "mdstat.c ($Revision: 1.5 $)";
static char	buf[BUFLEN+1]	= { 0, };
static char *	token		= NULL;
static int	lpos		= 0;
static int	tpos		= 0;
static size_t	count		= 0;

FILE *		mdstat_file	= NULL;
static int	mdstat_type	= MDSTAT_TYPE_WITHOUT_EVENTS;
struct pollfd	pfd;

int read_line(void) {
	size_t	bufstart;
	size_t	n;

	lpos++;
	if(lpos < count) {
		bufstart = count - lpos;
		for(n = 0; n < bufstart; n++, lpos++)
			buf[n] = buf[lpos];
		count = fread(buf+bufstart, 1, BUFLEN-bufstart, mdstat_file);
		if(count >= 0)
			count += bufstart;
	} else {
		count = fread(buf, 1, BUFLEN, mdstat_file);
	}
	if(count < 0) {
		ERR("Error in read_line: %d\n", errno);
		return -1;
	}

	for(lpos = 0; lpos < count; lpos++) {
		if(buf[lpos] == '\n') {
			break;
		}
	}
	buf[lpos] = '\0';

	return lpos;
}

int read_token(void) {
	int	rc;
	int	brace;

	while(tpos >= lpos) {
		if((rc = read_line()) <= 0)
			return rc;

		tpos = 0;
		/* Skip whitespace at beginning of line. */
		for(; tpos < lpos; tpos++) {
			if(buf[tpos] != ' ' && buf[tpos] != '\t')
				break;
		}
	}

	rc = brace = 0;
	token = buf + tpos;

	for(; tpos < lpos; tpos++, rc++) {
		if(buf[tpos] == '[')
			brace++;
		if(buf[tpos] == ']')
			brace--;
		if(!brace && (buf[tpos] == ' ' || buf[tpos] == '\t')) {
			buf[tpos] = '\0';
			tpos++;
			break;
		}
	}

	/* Skip any trailing whitespace. */
	for(; tpos < lpos; tpos++) {
		if(buf[tpos] != ' ' && buf[tpos] != '\t')
			break;
	}

	return rc;
}

void push_token(void) {
	int i;

	for(i = 0; buf[i] != '\0'; i++);
	buf[i] = ' ';
	tpos = token - buf;
}

long token2long(int base) {
	long	n;
	char	c;

	n = 0L;
	while(1) {
		c = tolower(*token);
		if(c >= '0' && c <= '9') {
			n = (n * base) + (c - '0');
		} else if(c >= 'a' && c <= 'f') {
			n = (n * base) + (c - 'a' + 10);
		} else {
			break;
		}
		token++;
	}

	return n;
}

int parse_device(struct mds_info *mdi, int md_nr) {
	static struct stat	st;
	static char		devpath[80];
	char *			s;
	int			major;
	int			minor;
	int			disk_nr;

	read_token();
	s = token;

	/*
	 * A device entry is either "[dev <major>:<minor>][<disk_nr>]" or
	 * "<devname>[<disk_nr>]".
	 */
	major = minor = 0;
	if(strncmp(token, "[dev ", 5) == 0) {
		token += 5;
		major = (int) token2long(16);
		token++;
		minor = (int) token2long(16);
		token++;
	} else {
		for(; *token != '[' && *token != '\0'; token++);
		if(*token == '[') {
			*token = '\0';
			sprintf(devpath, "%s/%s", DEVDIR, s);
			*token = '[';

			if(stat(devpath, &st) == 0 && S_ISBLK(st.st_mode)) {
				major = MAJOR(st.st_rdev);
				minor = MINOR(st.st_rdev);
			}
		}
	}

	/* A device entry ends with "[<nr>]" */
	if(*token != '[') {
		token = s;
		push_token();
		return -1;
	}
	token++;
	disk_nr = token2long(10);
	mdi->md[md_nr].disk[disk_nr].major = major;
	mdi->md[md_nr].disk[disk_nr].minor = minor;
	token++;

	/* The disk may be "(F)"ailed. */
	if(strcmp(token, "(F)") == 0) {
		mdi->md[md_nr].disk[disk_nr].status = disk_failed;
	} else {
		mdi->md[md_nr].disk[disk_nr].status = disk_active;
	}

	return (int) disk_nr;
}

int parse_array(struct mds_info *mdi) {
	int	rc;
	int	md_nr;
	int	seq_nr;
	
	token += 2;
	md_nr = (int) token2long(10);
	mdi->md[md_nr].nr = md_nr;

	/* Skip the colon, if it's there. */
	rc = read_token();
	if(strcmp(token, ":") == 0) {
		rc = read_token();
	}

	/* Is the array active? */
	if(strcmp(token, "active") == 0) {
		mdi->md[md_nr].status = array_active;
	}
	rc = read_token();

	/* What discipline is used */
	if(strcmp(token, "raid0") == 0) {
		mdi->md[md_nr].level = raid0;
	} else if(strcmp(token, "raid1") == 0) {
		mdi->md[md_nr].level = raid1;
	} else if(strcmp(token, "raid4") == 0) {
		mdi->md[md_nr].level = raid4;
	} else if(strcmp(token, "raid5") == 0) {
		mdi->md[md_nr].level = raid5;
	} else if(strcmp(token, "linear") == 0) {
		mdi->md[md_nr].level = linear;
	} else if(strcmp(token, "multipath") == 0) {
		mdi->md[md_nr].level = multipath;
	} else {
		ERR("md%d: Unknown discipline <%s>\n", md_nr, token);
		return -1;
	}

	/* Now get the devices. */
	seq_nr = 0;
	do {
		rc = parse_device(mdi, md_nr);
		if(rc >= 0) {
			mdi->md[md_nr].disks++;
			mdi->md[md_nr].disk[rc].seq_nr = seq_nr;
			seq_nr++;
			if(mdi->md[md_nr].disk[rc].status == disk_failed)
				mdi->md[md_nr].failed++;
		}
	} while(rc >= 0);

	return 0;
}

int mdstat_register(int type) {
	(void) revision;

	/* Already registered. */
	if(mdstat_file != NULL)
		return -1;

	if(type == MDSTAT_TYPE_WITH_EVENTS) {
		if((mdstat_file = fopen("/proc/mdstat", "r+")) == 0)
			return -1;
		mdstat_type = type;
	} else {
		if((mdstat_file = fopen("/proc/mdstat", "r")) == 0)
			return -1;
		mdstat_type = MDSTAT_TYPE_WITHOUT_EVENTS;
	}

	pfd.fd     = fileno(mdstat_file);
	pfd.events = POLLPRI;

	return 0;
}

int mdstat_unregister(void) {
	if(mdstat_file == NULL)
		return -1;

	fclose(mdstat_file);
	mdstat_file = NULL;

	return 0;
}

int mdstat_get(struct mds_info *mdi) {
	int	rc;

	memset(&mdi->md, 0, sizeof(mdi->md));
	mdi->events = -1;

	rewind(mdstat_file);

	count = lpos = tpos = 0;
	while((rc = read_token()) > 0) {
		if(strcmp(token, "Event:") == 0) {
			read_token();
			mdi->events = token2long(10);
		} else if(strncmp(token, "md", 2) == 0) {
			parse_array(mdi);
		}
	}

	return rc;
}

int mdstat_wait(int timeout) {

	if(mdstat_type == MDSTAT_TYPE_WITHOUT_EVENTS) {
		sleep(timeout/1000);
		return 0;
	} else {
		pfd.revents = 0;
		if(poll(&pfd, 1, timeout) > 0) {
			if(pfd.revents & POLLPRI)
				return 0;
		}

		return -1;
	}
}
